#
# Copyright (c) 2004 Brian Sutherland
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import os
import unittest
from StringIO import StringIO

import libxml2

from minittml import ttmldom
from minittml import data
from minittml.tests.helpers import stringify
from minittml.tests.helpers import verifyStrings


class TestWriteRead(unittest.TestCase):
    """Test by reading a TTML file and then re-writing it."""

    def __init__(self, infile):
        unittest.TestCase.__init__(self)
        self.test_id = os.path.basename('Read and Write Test: ' + infile)
        self.filename = infile

    def id(self):
        return self.test_id

    def runTest(self):
        """Test the reading and writing of a complete document."""
        # Read the document
        factory = ttmldom.DocumentFactory()
        readdoc = factory.parse(self.filename)
        objects = []
        objects.append(readdoc.popInputDescriptive())
        objects.append(readdoc.popVariableSet())
        objects.append(readdoc.popTimeDomain())
        while True:
            try:
                objects.append(readdoc.popDomain())
            except IndexError:
                break
        try:
            objects.append(readdoc.popDomainsvar())
        except IndexError:
            pass
        while True:
            try:
                objects.append(readdoc.popRootClassifier())
            except IndexError:
                break
        while True:
            try:
                objects.append(readdoc.popSoftConstraint())
            except IndexError:
                break
        while True:
            try:
                objects.append(readdoc.popHardConstraint())
            except IndexError:
                break
        # XXX - Check that the read document is empty!!!
        # Write the document
        writedoc = factory.create()
        for object in objects:
            writedoc.addObject(object)
        # Print both to strings
        writedocstring = stringify(writedoc.doc)
        readfile = open(self.filename)
        readdocstring = readfile.read()
        readfile.close()
        # Finally compare the written and read objects
        self.assert_(verifyStrings(writedocstring, readdocstring))


class TestSchema(unittest.TestCase):
    """Functional testing of the schema by running the test XML through it.

    This runs valid TTML snippets through the schema to make sure they pass.
    """

    schema = 'src/minittml/schema.rng'

    def __init__(self, infile):
        unittest.TestCase.__init__(self)
        self.test_id = os.path.basename('Test if Valid: ' + infile)
        self.filename = infile

    def id(self):
        return self.test_id

    def runTest(self):
        parser = libxml2.readerForFile(self.filename, None, 0)
        parser.RelaxNGValidate(self.schema)
        while parser.Read() == 1:
            self.assert_(parser.IsValid())


class TestSchemaInvalid(unittest.TestCase):
    """Functional testing of the schema by running the test XML through it.

    This runs invalid TTML snippets through the schema to make sure they fail.
    """

    schema = 'src/minittml/schema.rng'

    def __init__(self, infile):
        unittest.TestCase.__init__(self)
        self.test_id = os.path.basename('Test if Invalid: ' + infile)
        self.filename = infile

    def id(self):
        return self.test_id

    def runTest(self):
        parser = libxml2.readerForFile(self.filename, None,
                libxml2.XML_PARSE_NOWARNING +
                libxml2.XML_PARSE_NOERROR)
        parser.RelaxNGValidate(self.schema)
        tmp = StringIO()
        while parser.Read == 1 and parser.IsValid():
            pass
        self.assert_(not parser.IsValid())


def find_files(subdir):
    dirname = os.path.join(os.path.dirname(__file__), subdir)
    files = [os.path.join(dirname, fn) for fn in os.listdir(dirname)\
            if fn.endswith('.xml')]
    return files

def test_suite():
    suite = unittest.TestSuite()
    for doc in find_files('valid'):
        suite.addTest(TestWriteRead(doc))
        suite.addTest(TestSchema(doc))
    for doc in find_files('invalid'):
        suite.addTest(TestSchemaInvalid(doc))
    return suite
