#
# Copyright (c) 2004 Brian Sutherland <jinty@web.de>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

import unittest
from StringIO import StringIO
from sets import Set, ImmutableSet
from datetime import timedelta
from datetime import datetime
from datetime import MINYEAR
from xml.dom import minidom
from xml.dom import expatbuilder

from zope.interface import implements
from zope.interface.verify import verifyObject

from minittml import interfaces
from minittml import data
from minittml import ttmldom
from minittml.tests.helpers import stringify
from minittml.tests.helpers import verifyStrings

implementation = ttmldom.DOMImplementation()
expatbuilder.theDOMImplementation = implementation

def parse(str):
    """Parses a string into a DOM tree."""
    builder = expatbuilder.ExpatBuilder()
    return builder.parseString(str)


class Stub:
    pass


class ElementStub:
    """Stub for Elements"""

    def __init__(self, tagName, namespaceURI=None, prefix=None):
        self.tagName = tagName
        self.namespaceURI = namespaceURI
        self.prefix = prefix


class ExpatBuilderStub:

    def parseFile(self, file):
        file.parsed = True
        return file


class DocumentWrapperStub:

    def __init__(self, doc):
        self.doc = doc


class FileOpenStub:

    def __init__(self, file, access):
        self.name = file
        self.access = access
        self.parsed = False
        self.closed = False

    def close(self):
        self.closed = True


class ImplementationStub:

    def createDocumentType(self, qualifiedName, publicId, systemId):
        doctype = Stub()
        doctype.qualifiedName = qualifiedName
        doctype.publicId = publicId
        doctype.systemId = systemId
        return doctype

    def createDocument(self, namespaceURI, qualifiedName, doctype):
        self.namespaceURI = namespaceURI
        self.qualifiedName = qualifiedName
        self.doctype = doctype
        return self

#
# Tests
#

class TestHelperFunctions(unittest.TestCase):
    """Test the helper functions in the module not attached to classes"""

    def setUp(self):
        factory = ttmldom.DocumentFactory()
        self.doc = factory.create().doc

    def createSetOrVectorNode(self, doc, setname, tagName, elements):
        setdecl = doc.createElement('declare')
        name = doc.createElement('ci')
        name.appendChild(doc.createTextNode(setname))
        setdecl.appendChild(name)
        setnode = doc.createElement(tagName)
        for element in elements:
            node = doc.createElement('ci')
            node.appendChild(element)
            setnode.appendChild(node)
        setdecl.appendChild(setnode)
        return setdecl

    def test_getText(self):
        doc = self.doc
        element = doc.createElement('test')
        textnode = doc.createTextNode('Some Text')
        element.appendChild(textnode)
        self.assertEquals(ttmldom.getText(element), 'Some Text')
        textnode = doc.createTextNode(', and some more.')
        element.appendChild(textnode)
        self.assertEquals(ttmldom.getText(element),
                'Some Text, and some more.')


class TestBaseElement(unittest.TestCase):
    """Test the BaseElement data object."""

    def testremoveAll(self):
        instr = """
        <input-data>
            <test/>
            <testing/>
            <test/>
        </input-data>"""
        outstr = """
        <input-data>
            <testing/>
        </input-data>"""
        doc = parse(instr)
        doc.documentElement.removeAll('test')
        self.assert_(verifyStrings(outstr, stringify(doc.documentElement)))

    def testgetChildByTagName(self):
        instr = """
        <input-data>
            <test>yyy</test>
            <testing/>
            <test2/>
            <test>xxx</test>
        </input-data>"""
        doc = parse(instr)
        iter = doc.documentElement.getChildByTagName('test', 'test2')
        self.assertEquals(ttmldom.getText(iter.next()), 'yyy')
        self.assertEquals(ttmldom.getText(iter.next()), '')
        self.assertEquals(ttmldom.getText(iter.next()), 'xxx')
        self.assertRaises(StopIteration, iter.next)

    def testgetFirstChildByTagName(self):
        instr = """
        <input-data>
            <test>yyy</test>
            <testing/>
            <test2>zzz</test2>
            <test>xxx</test>
        </input-data>"""
        doc = parse(instr)
        node = doc.documentElement.getFirstChildByTagName('test', 'test2')
        self.assertEquals(ttmldom.getText(node), 'yyy')
        node = doc.documentElement.getFirstChildByTagName('test2')
        self.assertRaises(IndexError,
                doc.documentElement.getFirstChildByTagName, 'test3')

    def testorderedInsert(self):
        instr = """<input-data/>"""
        outstr1 = """
        <input-data>
            <test1/>
            <test3/>
            <test3/>
            <test2/>
        </input-data>"""
        outstr2 = """
        <input-data>
            <test1/>
            <test3/>
            <test2/>
        </input-data>"""
        doc = parse(instr)
        element = doc.documentElement
        test1 = doc.createElement('test1')
        self.assertRaises(ValueError, element.orderedInsert,
                test1)
        element.order = [ 'test1', 'test3', 'test2' ]
        test2 = doc.createElement('test2')
        test3a = doc.createElement('test3')
        test3b = doc.createElement('test3')
        element.orderedInsert(test3a)
        element.orderedInsert(test1)
        element.orderedInsert(test2)
        element.orderedInsert(test3b, onlyone=False)
        self.assert_(verifyStrings(outstr1, stringify(element)))
        test3c = doc.createElement('test3')
        element.orderedInsert(test3c)
        self.assert_(verifyStrings(outstr2, stringify(element)))


class TestDescElementMixin(unittest.TestCase):
    """Test the DescElement data object."""

    def setUp(self):
        factory = ttmldom.DocumentFactory()
        self.doc = factory.create().doc
        class Element(ttmldom.BaseElement, ttmldom.DescElementMixin):
            pass
        self.element = Element('input-data')
        self.element.ownerDocument = self.doc
        self.element.order = [ 'aref', 'author', 'desc' ]

    def testsetDesc(self):
        # XXX - convert to teting by string comparison
        self.element.setDesc('dummy')
        self.element.setDesc('Imagine something here')
        self.assertEquals(self.element.tagName, 'input-data')
        self.assertEquals(self.element.childNodes[0].tagName, 'desc')
        self.assertEquals(len(self.element.childNodes), 1)
        self.assertEquals(self.element.childNodes[0].childNodes[0].data,
                'Imagine something here')
        self.assertEquals(len(self.element.childNodes[0].childNodes), 1)

    def testreadDesc(self):
        # XXX - convert to teting by string comparison
        doc = self.doc
        garbage = doc.createElement('garbage')
        self.element.appendChild(garbage)
        self.assertEquals(self.element.readDesc(), '')
        node = doc.createElement('desc')
        textnode = doc.createTextNode('Somedesc')
        node.appendChild(textnode)
        self.element.appendChild(node)
        self.assertEquals(self.element.readDesc(), 'Somedesc')
        self.assertEquals(self.element.childNodes[1].firstChild.data,
                'Somedesc')

    def testsetAuthors(self):
        # XXX - convert to teting by string comparison
        self.element.setAuthors(['Original'])
        self.element.setAuthors(['Me.', u'you'])
        self.assertEquals(self.element.tagName, 'input-data')
        self.assertEquals(self.element.childNodes[0].tagName, 'author')
        self.assertEquals(self.element.childNodes[1].tagName, 'author')
        self.assertEquals(len(self.element.childNodes), 2)
        self.assertEquals(self.element.childNodes[0].childNodes[0].data,
                'Me.')
        self.assertEquals(len(self.element.childNodes[0].childNodes), 1)
        self.assertEquals(self.element.childNodes[1].childNodes[0].data,
                'you')
        self.assertEquals(len(self.element.childNodes[1].childNodes), 1)

    def testreadAuthors(self):
        # XXX - convert to teting by string comparison
        doc = self.doc
        garbage = doc.createElement('garbage')
        self.element.appendChild(garbage)
        self.assertEquals(self.element.readAuthors(), [])
        authornode = doc.createElement('author')
        authortext = doc.createTextNode('Somebody')
        authornode.appendChild(authortext)
        self.element.appendChild(authornode)
        authornode = doc.createElement('author')
        authortext = doc.createTextNode('Somebody Else')
        authornode.appendChild(authortext)
        self.element.appendChild(authornode)
        self.assertEquals(self.element.readAuthors(), ['Somebody',
                'Somebody Else'])
        self.assertEquals(self.element.childNodes[1].firstChild.data,
                'Somebody')

    def testReadAndSetRefs(self):
        # for input-data
        instr = """<input-data/>"""
        testobj1 = ['Me']
        outstr1 = """
        <input-data>
            <aref>
                <li>Me</li>
            </aref>
        </input-data>"""
        doc = parse(instr)
        doc.documentElement.setRefs(testobj1)
        self.assert_(verifyStrings(outstr1, stringify(doc.documentElement)))
        obj = doc.documentElement.readRefs()
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(outstr1, stringify(doc.documentElement)))
        testobj2 = ['One', u'Two']
        outstr2 = """
        <input-data>
            <aref>
                <li>One</li>
                <li>Two</li>
            </aref>
        </input-data>"""
        doc.documentElement.setRefs(testobj2)
        self.assert_(verifyStrings(outstr2, stringify(doc.documentElement)))
        obj = doc.documentElement.readRefs()
        self.assertEquals(obj, testobj2)
        self.assert_(verifyStrings(outstr2, stringify(doc.documentElement)))
        # Result Node
        instr = """<result/>"""
        testobj = ['Me']
        outstr = """
        <result>
            <pref>
                <li>Me</li>
            </pref>
        </result>"""
        doc = parse(instr)
        doc.documentElement.setRefs(testobj)
        self.assert_(verifyStrings(outstr, stringify(doc.documentElement)))
        obj = doc.documentElement.readRefs()
        self.assertEquals(obj, testobj)
        self.assert_(verifyStrings(outstr, stringify(doc.documentElement)))
        # other
        element = ttmldom.InputElement('other')
        element.ownerDocument = doc
        self.assertRaises(TypeError, element.setRefs, testobj)
        self.assertRaises(TypeError, element.readRefs)

    def testsetDescriptive(self):
        authors = ['Bukowski', 'Gutierrez']
        desc = 'Similarities.'
        refs = [u'Ham on Rye', 'Anclado en la tierra de nadie']
        desc_obj = data.Descriptive(authors, desc, refs)
        self.element.setDescriptive(desc_obj)
        self.assertEquals(self.element.readDesc(), desc)
        self.assertEquals(self.element.readAuthors(), authors)
        self.assertEquals(self.element.readRefs(), refs)

    def testpopDescriptive(self):
        authors = ImmutableSet(['Bukowski', 'Gutierrez'])
        desc = 'Similarities.'
        refs = ImmutableSet([u'Ham on Rye', 'Anclado en la tierra de nadie'])
        self.element.setAuthors(authors)
        self.element.setDesc(desc)
        self.element.setRefs(refs)
        desc_obj = self.element.popDescriptive()
        self.assertEquals(desc_obj.getDesc(), desc)
        self.assertEquals(desc_obj.getAuthors(), authors)
        self.assertEquals(desc_obj.getRefs(), refs)
        verifyObject(interfaces.IInputDescription, desc_obj)
        self.assertRaises(IndexError,
                self.element.getElementsByTagName('aref').pop)
        self.assertRaises(IndexError,
                self.element.getElementsByTagName('author').pop)
        self.assertRaises(IndexError,
                self.element.getElementsByTagName('desc').pop)


class TestVariablesElement(unittest.TestCase):

    def testpopiAndSetVariableSet(self):
        instr = "\n<variables/>"
        prefix = "\n<variables>"
        suffex = "\n</variables>\n"
        attrset_name = 'classrooms'
        members = ('A100', 'B101', 'B103', 'A201')
        attr_names = ('capacity', )
        attrs = (('50', '50', '50', '30'), )
        element_attr_names = ('duration', )
        element_attrs = (('50', '50', '', None), )
        testobj1 = data.VariableSet(attrset_name, members,
                attribute_names=attr_names, attributes=attrs,
                element_attr_names=element_attr_names,
                element_attrs=element_attrs)
        outstr1 = """
        <attrset>
            <declare>
                <ci>classrooms</ci>
                <vector>
                    <ci duration='50'>A100</ci>
                    <ci duration='50'>B101</ci>
                    <ci duration=''>B103</ci>
                    <ci>A201</ci>
                </vector>
            </declare>
            <declare>
                <ci>capacity</ci>
                <vector>
                    <ci>50</ci>
                    <ci>50</ci>
                    <ci>50</ci>
                    <ci>30</ci>
                </vector>
            </declare>
        </attrset>"""
        attrset_name = 'simpledomain'
        testobj2 = data.VariableSet(attrset_name, members,
                element_attr_names=element_attr_names,
                element_attrs=element_attrs)
        outstr2 = """
        <declare>
            <ci>simpledomain</ci>
            <set>
                <ci duration='50'>A100</ci>
                <ci duration='50'>B101</ci>
                <ci duration=''>B103</ci>
                <ci>A201</ci>
            </set>
        </declare>"""
        doc = parse(instr)
        doc.documentElement.setVariableSet(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        doc.documentElement.setVariableSet(testobj2)
        verifyStrings(prefix + outstr2 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popVariableSet()
        verifyObject(interfaces.IVariableSet, obj)
        self.assertEquals(obj, testobj2)
        verifyStrings(instr,
                stringify(doc.documentElement))
        doc.documentElement.setVariableSet(testobj1)
        obj = doc.documentElement.popVariableSet()
        verifyObject(interfaces.IVariableSet, obj)
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))
        self.assertRaises(IndexError, doc.documentElement.popVariableSet)


class TestDomainsElement(unittest.TestCase):

    def testPopAndSetDomain(self):
        instr = "\n<domains/>"
        prefix = "\n<domains>"
        suffex = "\n</domains>\n"
        attrset_name = 'classrooms'
        members = ('A100', 'B101', 'B103', 'A201')
        attr_names = ('capacity', )
        attrs = (('50', '50', '50', '30'), )
        testobj1 = data.Domain(attrset_name, members,
                attribute_names=attr_names, attributes=attrs)
        outstr1 = """
        <attrset>
            <declare>
                <ci>classrooms</ci>
                <vector>
                    <ci>A100</ci>
                    <ci>B101</ci>
                    <ci>B103</ci>
                    <ci>A201</ci>
                </vector>
            </declare>
            <declare>
                <ci>capacity</ci>
                <vector>
                    <ci>50</ci>
                    <ci>50</ci>
                    <ci>50</ci>
                    <ci>30</ci>
                </vector>
            </declare>
        </attrset>"""
        attrset_name = 'simpledomain'
        testobj2 = data.Domain(attrset_name, members)
        outstr2 = """
        <declare>
            <ci>simpledomain</ci>
            <set>
                <ci>A100</ci>
                <ci>B101</ci>
                <ci>B103</ci>
                <ci>A201</ci>
            </set>
        </declare>"""
        doc = parse(instr)
        doc.documentElement.appendDomain(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        doc.documentElement.appendDomain(testobj2)
        verifyStrings(prefix + outstr1 + outstr2 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popDomain()
        verifyObject(interfaces.IDomain, obj)
        self.assertEquals(obj, testobj1)
        verifyStrings(prefix + outstr2 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popDomain()
        verifyObject(interfaces.IDomain, obj)
        self.assertEquals(obj, testobj2)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))
        self.assertRaises(IndexError, doc.documentElement.popDomain)

    def testPopAndSetDomainsvar(self):
        instr = """\n<domains/>\n"""
        testobj = data.Domainsvar('R',
                ['T', 'classrooms'])
        outstr = """
        <domains>
            <domainsvar>
                <declare>
                    <ci>R</ci>
                    <apply>
                        <cartesianproduct/>
                        <ci>T</ci>
                        <ci>classrooms</ci>
                    </apply>
                </declare>
            </domainsvar>
        </domains>"""
        doc = parse(instr)
        doc.documentElement.setDomainsvar(testobj)
        self.assert_(verifyStrings(outstr, stringify(doc.documentElement)))
        obj = doc.documentElement.popDomainsvar()
        verifyObject(interfaces.IDomainsvar, obj)
        self.assertEquals(obj, testobj)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))
        self.assertRaises(TypeError,
                doc.documentElement.setDomainsvar, 's')


class TestTimeElement(unittest.TestCase):

    def testconvertTimeStrToObj(self):
        instr = """
        <time/>"""
        element = parse(instr).documentElement
        strs = ['1984-2-14T15:34:00']
        objs = [datetime(1984, 2, 14, 15, 34)]
        for str, obj in zip(strs, objs):
            self.assertEquals(element.convertTimeStrToObj(str), obj)

    def testconvertDurationStrToObj(self):
        instr = """
        <time/>"""
        element = parse(instr).documentElement
        strs = ['PT1H', 'P3DT4H5M6S']
        objs = [timedelta(hours=1),
                timedelta(days=3, hours=4, minutes=5,
                        seconds=6)]
        for str, obj in zip(strs, objs):
            self.assertEquals(element.convertDurationStrToObj(str),
                    obj)
        strs = ['P3MT1H', 'P4Y3DT4H5M6S']
        for str in strs:
            self.assertRaises(ValueError, element.convertDurationStrToObj, str)

    def testSetAndReadTimeDomainBasic(self):
        instr = """<time/>"""
        intervals = [
                (datetime(1984,4,13,hour=8), timedelta(minutes=50)),
                (datetime(1984,4,13,hour=9), timedelta(minutes=50)),
                (datetime(1984,4,14,hour=8), timedelta(minutes=50)),
                (datetime(1984,4,14,hour=9), timedelta(minutes=50))]
        testtimeobj = data.TimeDomain('T', intervals, True)
        outstr = """
        <time>
            <declare>
                <ci>T</ci>
                <set interval='discrete'>
                    <cn>1984-04-13T08:00:00<sep/>P0DT3000S</cn>
                    <cn>1984-04-13T09:00:00<sep/>P0DT3000S</cn>
                    <cn>1984-04-14T08:00:00<sep/>P0DT3000S</cn>
                    <cn>1984-04-14T09:00:00<sep/>P0DT3000S</cn>
                </set>
            </declare>
        </time>"""
        doc = parse(instr)
        doc.documentElement.setTimeDomainBasic(testtimeobj)
        self.assert_(verifyStrings(outstr, stringify(doc.documentElement)))
        timeobj = doc.documentElement.readTimeDomainBasic()
        verifyObject(interfaces.ITimeDomain, timeobj)
        self.assertEquals(timeobj, testtimeobj)
        self.assert_(verifyStrings(outstr, stringify(doc.documentElement)))

    def testSetAndPopTimeDomain(self):
        instr = """
        <time>
        </time>"""
        testtimeobjs = []
        # basic
        intervals = [
                (datetime(1984,4,13,hour=8), timedelta(minutes=50)),
                (datetime(1984,4,13,hour=9), timedelta(minutes=50)),
                (datetime(1984,4,14,hour=8), timedelta(minutes=50)),
                (datetime(1984,4,14,hour=9), timedelta(minutes=50))]
        testtimeobjs.append(data.TimeDomain('Time', intervals, False))
        # tmatrix
        testtimeobjs.append(data.TmatrixTimeDomain([2, 4]))
        # test all the objs with the functions
        for testtimeobj in testtimeobjs:
            doc = parse(instr)
            doc.documentElement.setTimeDomain(testtimeobj)
            outstr = stringify(doc.documentElement)
            testdoc = parse(outstr)
            timeobj = testdoc.documentElement.popTimeDomain()
            self.assertEquals(testtimeobj, timeobj)
            self.assert_(verifyStrings(instr, stringify(testdoc.documentElement)))


class TestConstraintsElement(unittest.TestCase):

    def testNotSameConstraint(self):
        instr = """\n<hard/>"""
        prefix = "\n<hard>"
        suffex = "\n</hard>\n"
        testobj1 = data.NotSameConstraint(['CSE211.01', 'A100'],
                attributes={'dim': 'classroom'}, name='const1')
        outstr1 = """
                <coi>
                    <declare>
                        <ci>const1</ci>
                        <apply>
                            <notsame dim='classroom'/>
                            <ci>CSE211.01</ci>
                            <ci>A100</ci>
                        </apply>
                    </declare>
                </coi>"""
        doc = parse(instr)
        doc.documentElement.appendConstraint(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.INotSameConstraint, obj)
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))

    def testNoOverlapConstraint(self):
        instr = """\n<hard/>"""
        prefix = "\n<hard>"
        suffex = "\n</hard>\n"
        testobj1 = data.NoOverlapConstraint(['CSE211.01', 'A100'],
                attributes={'dim': 'classroom'}, name='const1')
        outstr1 = """
                <coi>
                    <declare>
                        <ci>const1</ci>
                        <apply>
                            <nooverlap dim='classroom'/>
                            <ci>CSE211.01</ci>
                            <ci>A100</ci>
                        </apply>
                    </declare>
                </coi>"""
        doc = parse(instr)
        doc.documentElement.appendConstraint(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.INoOverlapConstraint, obj)
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))

    def testPresetConstraint(self):
        instr = """\n<hard/>"""
        prefix = "\n<hard>"
        suffex = "\n</hard>\n"
        testobj1 = data.PresetConstraint(['CSE211.01', 'A100'],
                attributes={'dim': 'classroom'}, name='const1')
        outstr1 = """
                <coi>
                    <declare>
                        <ci>const1</ci>
                        <apply>
                            <preset dim='classroom'/>
                            <ci>CSE211.01</ci>
                            <ci>A100</ci>
                        </apply>
                    </declare>
                </coi>"""
        doc = parse(instr)
        doc.documentElement.appendConstraint(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.IPresetConstraint, obj)
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))

    def testExcludeConstraint(self):
        instr = """\n<hard/>"""
        prefix = "\n<hard>"
        suffex = "\n</hard>\n"
        testobj1 = data.ExcludeConstraint(['CSE211.01', 'A100'],
                attributes={'dim': 'classroom'}, name='const1')
        outstr1 = """
                <coi>
                    <declare>
                        <ci>const1</ci>
                        <apply>
                            <exclude dim='classroom'/>
                            <ci>CSE211.01</ci>
                            <ci>A100</ci>
                        </apply>
                    </declare>
                </coi>"""
        doc = parse(instr)
        doc.documentElement.appendConstraint(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.IExcludeConstraint, obj)
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))

    def testOrderingConstraint(self):
        instr = """\n<hard/>"""
        prefix = "\n<hard>"
        suffex = "\n</hard>\n"
        testobj1 = data.OrderingConstraint(['CSE211.01', 'A100'],
                attributes={'dim': 'classroom'}, name='const1')
        outstr1 = """
                <coi>
                    <declare>
                        <ci>const1</ci>
                        <apply>
                            <ordering dim='classroom'/>
                            <ci>CSE211.01</ci>
                            <ci>A100</ci>
                        </apply>
                    </declare>
                </coi>"""
        doc = parse(instr)
        doc.documentElement.appendConstraint(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.IOrderingConstraint, obj)
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))

    def testEventSprConstraint(self):
        instr = """\n<hard/>"""
        prefix = "\n<hard>"
        suffex = "\n</hard>\n"
        testobj1 = data.EventSprConstraint(['CSE211.01', 'A100'],
                attributes={'dim': 'classroom'}, name='const1')
        outstr1 = """
                <coi>
                    <declare>
                        <ci>const1</ci>
                        <apply>
                            <eventspr dim='classroom'/>
                            <ci>CSE211.01</ci>
                            <ci>A100</ci>
                        </apply>
                    </declare>
                </coi>"""
        doc = parse(instr)
        doc.documentElement.appendConstraint(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.IEventSprConstraint, obj)
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))

    def testFreeSprConstraint(self):
        instr = """\n<hard/>"""
        prefix = "\n<hard>"
        suffex = "\n</hard>\n"
        testobj1 = data.FreeSprConstraint(['CSE211.01', 'A100'],
                attributes={'dim': 'classroom'}, name='const1')
        outstr1 = """
                <coi>
                    <declare>
                        <ci>const1</ci>
                        <apply>
                            <freespr dim='classroom'/>
                            <ci>CSE211.01</ci>
                            <ci>A100</ci>
                        </apply>
                    </declare>
                </coi>"""
        doc = parse(instr)
        doc.documentElement.appendConstraint(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.IFreeSprConstraint, obj)
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))

    def testAttrCompConstraint(self):
        instr = """\n<hard/>"""
        prefix = "\n<hard>"
        suffex = "\n</hard>\n"
        testobj1 = data.AttrCompConstraint(['CSE211.01', 'A100'],
                attributes={'dim': 'classroom'}, name='const1')
        outstr1 = """
                <coi>
                    <declare>
                        <ci>const1</ci>
                        <apply>
                            <attrcomp dim='classroom'/>
                            <ci>CSE211.01</ci>
                            <ci>A100</ci>
                        </apply>
                    </declare>
                </coi>"""
        doc = parse(instr)
        doc.documentElement.appendConstraint(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.IAttrCompConstraint, obj)
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))

    def testResNoClashConstraint(self):
        instr = """\n<hard/>"""
        prefix = "\n<hard>"
        suffex = "\n</hard>\n"
        testobj1 = data.ResNoClashConstraint(['CSE211.01', 'A100'],
                attributes={'dim': 'classroom'}, name='const1')
        outstr1 = """
                <coi>
                    <declare>
                        <ci>const1</ci>
                        <apply>
                            <resnoclash dim='classroom'/>
                            <ci>CSE211.01</ci>
                            <ci>A100</ci>
                        </apply>
                    </declare>
                </coi>"""
        doc = parse(instr)
        doc.documentElement.appendConstraint(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.IResNoClashConstraint, obj)
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))

    def testChksumConstraint(self):
        instr = """\n<hard/>"""
        prefix = "\n<hard>"
        suffex = "\n</hard>\n"
        testobj1 = data.ChksumConstraint(['CSE211.01', 'A100'],
                attributes={'dim': 'classroom'}, name='const1')
        outstr1 = """
                <coi>
                    <declare>
                        <ci>const1</ci>
                        <apply>
                            <chksum dim='classroom'/>
                            <ci>CSE211.01</ci>
                            <ci>A100</ci>
                        </apply>
                    </declare>
                </coi>"""
        doc = parse(instr)
        doc.documentElement.appendConstraint(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.IChksumConstraint, obj)
        self.assertEquals(obj, testobj1)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))

    def testPoppingConstraints(self):
        #Test a constraint
        instr = """\n<hard/>"""
        prefix = "\n<hard>"
        suffex = "\n</hard>\n"
        testobj1 = data.PresetConstraint(['CSE211.01', 'A100'],
                attributes={'dim': 'classroom'}, name='const1')
        outstr1 = """
                <coi>
                    <declare>
                        <ci>const1</ci>
                        <apply>
                            <preset dim='classroom'/>
                            <ci>CSE211.01</ci>
                            <ci>A100</ci>
                        </apply>
                    </declare>
                </coi>"""
        testobj2 = data.ExcludeConstraint(['Class a', 'A100'],
                {'dim': 'classroom'})
        outstr2 = """
                <coi>
                    <apply>
                        <exclude dim='classroom'/>
                        <ci>Class a</ci>
                        <ci>A100</ci>
                    </apply>
                </coi>"""
        doc = parse(instr)
        doc.documentElement.appendConstraint(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        doc.documentElement.appendConstraint(testobj2)
        verifyStrings(prefix + outstr1 + outstr2 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.IPresetConstraint, obj)
        self.assertEquals(obj, testobj1)
        verifyStrings(prefix + outstr2 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popConstraint()
        verifyObject(interfaces.IExcludeConstraint, obj)
        self.assertEquals(obj, testobj2)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))
        self.assertRaises(IndexError, doc.documentElement.popConstraint)


class TestClassifiersElement(unittest.TestCase):

    def testPopAndSetRootClassifier(self):
        instr = "\n<classifiers/>"
        prefix = "\n<classifiers>"
        suffex = "\n</classifiers>\n"
        members = ('A100', 'B101', 'B103', 'A201')
        testobj1 = data.Classifier('classrooms', members)
        outstr1 = """
        <rootcl>
            <declare>
                <ci>classrooms</ci>
                <set>
                    <ci>A100</ci>
                    <ci>B101</ci>
                    <ci>B103</ci>
                    <ci>A201</ci>
                </set>
            </declare>
        </rootcl>"""
        members = ('A100s ', 'B101s ', testobj1, 'A201s ')
        testobj2 = data.Classifier('simpledomain', members)
        outstr2 = """
        <rootcl>
            <declare>
                <ci>simpledomain</ci>
                <set>
                    <ci>A100s </ci>
                    <ci>B101s </ci>
                    <ci>
                        <declare>
                            <ci>classrooms</ci>
                            <set>
                                <ci>A100</ci>
                                <ci>B101</ci>
                                <ci>B103</ci>
                                <ci>A201</ci>
                            </set>
                        </declare>
                    </ci>
                    <ci>A201s </ci>
                </set>
            </declare>
        </rootcl>"""
        testobj3 = data.Classifier('simpledomain', [
                ('A100s ', {'attr': '1'}),
                ('B101s ', {}),
                (testobj1, {}),
                ('A201s ', {})])
        doc = parse(instr)
        doc.documentElement.appendRootClassifier(testobj1)
        verifyStrings(prefix + outstr1 + suffex,
                stringify(doc.documentElement))
        doc.documentElement.appendRootClassifier(testobj2)
        verifyStrings(prefix + outstr1 + outstr2 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popRootClassifier()
        verifyObject(interfaces.IClassifier, obj)
        self.assertEquals(obj, testobj1)
        verifyStrings(prefix + outstr2 + suffex,
                stringify(doc.documentElement))
        obj = doc.documentElement.popRootClassifier()
        verifyObject(interfaces.IClassifier, obj)
        self.assertEquals(obj, testobj2)
        self.assert_(verifyStrings(instr, stringify(doc.documentElement)))
        self.assertRaises(IndexError, doc.documentElement.popRootClassifier)
        self.assertRaises(TypeError, doc.documentElement.appendRootClassifier,
                testobj3)
        self.assertRaises(TypeError, doc.documentElement.appendRootClassifier,
                'sss')


class TestOutputElement(unittest.TestCase):

    def test(self):
        # Nothing to test - yet!
        pass


class TestTestResultsElement(unittest.TestCase):

    def test(self):
        # Nothing to test - yet!
        pass

class TestTimeTablingElement(unittest.TestCase):

    def test(self):
        # Nothing to test!
        pass


class TestDOMImplementation(unittest.TestCase):
    """Test the DOMImplementation data object."""

    def testDOMImplementation(self):
        impl = ttmldom.DOMImplementation()
        impl.documnet_hook = "Imaginary document"
        self.assert_(isinstance(impl._create_document(), ttmldom.Document))


class TestDocument(unittest.TestCase):
    """Test the Document data object."""

    def setUp(self):
        factory = ttmldom.DocumentFactory()
        self.doc = factory.create().doc

    def test_chooseElementType(self):
        doc = self.doc
        self.assertEquals(self.doc._chooseElementType('input-data'),
                ttmldom.InputElement)
        self.assertEquals(self.doc._chooseElementType('domains'),
                ttmldom.DomainsElement)
        self.assertEquals(self.doc._chooseElementType('constraints'),
                ttmldom.ConstraintsElement)
        self.assertEquals(self.doc._chooseElementType('classifiers'),
                ttmldom.ClassifiersElement)
        self.assertEquals(self.doc._chooseElementType('output'),
                ttmldom.OutputElement)
        self.assertEquals(self.doc._chooseElementType('time'),
                ttmldom.TimeDomainElement)
        self.assertEquals(self.doc._chooseElementType('test-results'),
                ttmldom.TestResultsElement)
        self.assertEquals(self.doc._chooseElementType('time-tabling'),
                ttmldom.TimeTablingElement)
        self.assertEquals(self.doc._chooseElementType('other'),
                ttmldom.BaseElement)

    def testcreateElement(self):
        doc = self.doc
        doc._chooseElementType = lambda x: ElementStub
        element = doc.createElement('Test')
        self.assertEquals(element.tagName, 'Test')
        self.assertEquals(element.namespaceURI, None)
        self.assertEquals(element.prefix, None)
        self.assertEquals(element.ownerDocument, doc)

    def testcreateElementNS(self):
        doc = self.doc
        doc._chooseElementType = lambda x: ElementStub
        element = doc.createElementNS('http://lulu.d/me', 'lulu:Test')
        self.assertEquals(element.tagName, 'lulu:Test')
        self.assertEquals(element.namespaceURI, 'http://lulu.d/me')
        self.assertEquals(element.prefix, 'lulu')
        self.assertEquals(element.ownerDocument, doc)


class TestDocumentWrapper(unittest.TestCase):
    """Test the DocumentWrapper data object."""

    def setUp(self):
        doctype = implementation.createDocumentType('time-tabling', '', '')
        self.doc = implementation.createDocument(
            'http://schema', 'time-tabling', doctype)
        self.helper = ttmldom.DocumentWrapper(self.doc)

    def test__init__(self):
        self.assertRaises(TypeError, ttmldom.DocumentWrapper, 'me')
        verifyObject(interfaces.IDoc, self.helper)
        self.assertEquals(self.helper.doc, self.doc)
        self.assertEquals(self.helper.doc_node, self.doc.documentElement)

    def testError(self):
        self.assertRaises(TypeError, self.helper.addObject, 'S')
        self.assertRaises(IndexError, self.helper.getMappedNode, 'hard')

    def testpopInputDescriptive(self):
        obj = data.InputDescription(['me'], '', [])
        self.helper.addObject(obj)
        self.assertEquals(obj, self.helper.popInputDescriptive())

    def testpopVariableSet(self):
        obj = data.VariableSet('Set Name', ('Element 1', 'Element 2'))
        self.helper.addObject(obj)
        self.assertEquals(obj, self.helper.popVariableSet())

    def testpopTimeDomain(self):
        obj = data.TimeDomain('Name',
                [(datetime(2000, 11, 1), timedelta(hours=1))], True)
        self.helper.addObject(obj)
        self.assertEquals(obj, self.helper.popTimeDomain())

    def testpopTmatrixTimeDomain(self):
        obj = data.TmatrixTimeDomain([2, 4], name='T')
        self.helper.addObject(obj)
        self.assertEquals(obj, self.helper.popTimeDomain())

    def testpopDomain(self):
        obj = data.Domain('Set Name', ('Element 1', 'Element 2'))
        self.helper.addObject(obj)
        self.assertEquals(obj, self.helper.popDomain())

    def testpopDomainsvar(self):
        obj = data.Domainsvar('domains', ['d1', 'd2'])
        self.helper.addObject(obj)
        self.assertEquals(obj, self.helper.popDomainsvar())

    def testpopRootClassifier(self):
        obj = data.Classifier('Set Name',
                ('Element 1', 'Element 2'))
        self.helper.addObject(obj)
        self.assertEquals(obj, self.helper.popRootClassifier())

    def testAppendAndPopConstraint(self):
        # Preset Constraint
        obj = data.PresetConstraint(['A100', 'classifier name'])
        self.helper.addObject(obj)
        self.assertEquals(obj, self.helper.popHardConstraint())
        self.assertRaises(IndexError, self.helper.popHardConstraint)
        # Exclude Constraint
        obj = data.ExcludeConstraint(['A100', 'classifier name'],
                {'dim': 'classrooms'}, ishard=False)
        self.helper.addObject(obj)
        self.assertEquals(obj, self.helper.popSoftConstraint())
        self.assertRaises(IndexError, self.helper.popSoftConstraint)


class TestDocumentFactory(unittest.TestCase):
    """Test the Document Factory."""

    def test(self):
        factory = ttmldom.DocumentFactory()
        verifyObject(interfaces.IDocumentFactory, factory)

    def test_parse(self):
        factory = ttmldom.DocumentFactory()
        file = StringIO("\n<classifiers/>\n")
        doc = factory.parse(file)
        self.assertEquals(file.getvalue(), stringify(doc.doc.documentElement))
        factory.builder_hook = ExpatBuilderStub
        factory.documentwrapper_hook = DocumentWrapperStub
        factory.fileopen_hook = FileOpenStub
        result = factory.parse('file')
        self.assert_(result.doc.parsed)
        self.assert_(result.doc.closed)
        self.assertEquals(result.doc.access, 'rb')
        self.assertEquals(result.doc.name, 'file')


    def test_create(self):
        factory = ttmldom.DocumentFactory()
        factory.documentwrapper_hook = DocumentWrapperStub
        factory.implementation_hook = ImplementationStub()
        result = factory.create(schema='bogus')
        self.assertEquals(result.doc.qualifiedName, 'time-tabling')
        self.assertEquals(result.doc.namespaceURI,
                'bogus')
        self.assertEquals(result.doc.doctype.qualifiedName, 'time-tabling')
        self.assertEquals(result.doc.doctype.systemId, '')
        self.assertEquals(result.doc.doctype.publicId, '')


def test_suite():
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestHelperFunctions))
    suite.addTest(unittest.makeSuite(TestDescElementMixin))
    suite.addTest(unittest.makeSuite(TestBaseElement))
    suite.addTest(unittest.makeSuite(TestVariablesElement))
    suite.addTest(unittest.makeSuite(TestDomainsElement))
    suite.addTest(unittest.makeSuite(TestTimeElement))
    suite.addTest(unittest.makeSuite(TestConstraintsElement))
    suite.addTest(unittest.makeSuite(TestClassifiersElement))
    suite.addTest(unittest.makeSuite(TestOutputElement))
    suite.addTest(unittest.makeSuite(TestTestResultsElement))
    suite.addTest(unittest.makeSuite(TestTimeTablingElement))
    suite.addTest(unittest.makeSuite(TestDOMImplementation))
    suite.addTest(unittest.makeSuite(TestDocument))
    suite.addTest(unittest.makeSuite(TestDocumentWrapper))
    suite.addTest(unittest.makeSuite(TestDocumentFactory))
    return suite
