# coding: utf-8
#
# Copyright (c) 2004 Brian Sutherland
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
"""
TTML parse/write library.

This library implements extensions to the python minidom classes which make it
easier to parse/write TTML documents.

This library only implements a sub-set of TTML, or more specifically the short
cut/common functions as implementing a full MathML parser was not necessary. A
schema is provided with the parser which can validate inputs to the parser. The
aim of this schema is not to completely describe TTML, but only to provide a
means of parser input checking.

The TTML format is taken, as accurately as possible, from the paper:
    E. zcan, Towards an XML based standard for Timetabling Problems: TTML,
    http://cse.yeditepe.edu.tr/~eozcan/research/TTML/

Please note that this paper is not and does not try to be a standard or
complete.

WARNING: A large proportion of these interfaces are unimplemented and subject
change.
"""

import datetime
from xml.dom import minidom
from xml.dom import expatbuilder
from sets import Set

from zope.interface import implements

import interfaces
import data

# XXX - for debugging
from xml.dom import ext

#
# DOM functions
#

def getText(node):
    return ''.join([subnode.data for subnode in node.childNodes\
            if subnode.nodeType == subnode.TEXT_NODE])

class BaseElement(minidom.Element):
    """A subclass of a DOM element with basic helper functions.

    This is a utility sub class of a DOM element to be further sub classed
    for specific node types within a TTML document.
    """

    order = []

    def getChildByTagName(self, *tagnames):
        """Returns a generator over all of the child elements with tagname.

        This provides sort of the same functionality as getElementsByTagName,
        but is not recursive and allows the specification of multiple tagnames.
        """
        for node in self.childNodes[:]:
            if node.nodeType == self.ELEMENT_NODE and\
                    node.tagName in tagnames:
                yield node

    def getFirstChildByTagName(self, *tagnames):
        """Get the first child element with TagName in tagnames.

        A convienience method which duplicates the functionality of
        getChildByTagName somewhat, but makes the code more beautiful.
        """
        for node in self.childNodes:
            if node.nodeType == self.ELEMENT_NODE:
                if node.tagName in tagnames:
                    return node
        raise IndexError("No child elements with such tagnames")

    def removeAll(self, *tagNames):
        for node in self.getChildByTagName(*tagNames):
            self.removeChild(node)

    def orderedInsert(self, node, onlyone=True):
        # Check if the node falls in our ordering scheme
        if node.tagName not in self.order:
            raise ValueError("%s not a sub-element of %s." %
                    (node.tagName, self.tagName))
        # If there should only be one of this node-type
        if onlyone:
            self.removeAll(node.tagName)
        # Get a list of nodes previous to this one
        prev = []
        for i in self.order:
            prev.append(i)
            if node.tagName == i:
                break
        # Insert the node after any listed as previous
        for i in self.childNodes:
            if i.nodeType != i.ELEMENT_NODE:
                continue
            if i.tagName not in prev:
                self.insertBefore(node, i)
                return node
        self.appendChild(node)
        return node


class DescElementMixin:
    """A mixin for adding descriptive information functions to a TTML element.

    This mixin adds the functions setDescriptive/getDescriptive to a dom node.
    These functions handle objects with the IDescriptive interface.
    """

    def setDesc(self, desc):
        if not desc:
            self.removeAll('desc')
            return
        doc = self.ownerDocument
        descnode = doc.createElement('desc')
        descnode.appendChild(doc.createTextNode(desc))
        self.orderedInsert(descnode)

    def readDesc(self):
        for node in self.getElementsByTagName('desc'):
            return getText(node)
        return ''

    def setAuthors(self, authors):
        self.removeAll('author')
        doc = self.ownerDocument
        for author in authors:
            authornode = doc.createElement('author')
            authornode.appendChild(doc.createTextNode(author))
            self.orderedInsert(authornode, onlyone=False)

    def readAuthors(self):
        authors = []
        for node in self.getElementsByTagName('author'):
            author = getText(node)
            if author:
                authors.append(author)
        return authors

    def setRefs(self, refs):
        # Reference section has different names in the input or result, sigh
        if self.tagName == 'result':
            tagName = 'pref'
        elif self.tagName == 'input-data':
            tagName = 'aref'
        else:
            raise TypeError("Unknown parent node TagName")
        # Don't put a refs section if not wanted
        if not refs:
            self.removeAll(tagName)
            return
        # get the parent doc
        doc = self.ownerDocument
        # Put the refs onto the refnode
        refnode = doc.createElement(tagName)
        for ref in refs:
            linode = doc.createElement('li')
            linode.appendChild(doc.createTextNode(ref))
            refnode.appendChild(linode)
        # Put the refnode onto the parent node
        self.orderedInsert(refnode)

    def readRefs(self):
        if self.tagName == 'result':
            tagName = 'pref'
        elif self.tagName == 'input-data':
            tagName = 'aref'
        else:
            raise TypeError("Unknown parent node TagName")
        return [getText(node) for node in self.getElementsByTagName('li')]

    def setDescriptive(self, desc):
        self.setAuthors(desc.getAuthors())
        self.setDesc(desc.getDesc())
        self.setRefs(desc.getRefs())

    def popDescriptive(self):
        if self.tagName == 'input-data':
            desc =  data.InputDescription(self.readAuthors(), self.readDesc(),
                    self.readRefs())
        self.removeAll('author', 'pref', 'aref', 'desc')
        return desc


class FunctionMixin(BaseElement):
    """Give a node the ability to handle function objects."""

    functionMap = {}

    def putName(self, function):
        """Do the name setting part of a funcrion node"""
        doc = self.ownerDocument
        if function.getName():
            topnode = doc.createElement('declare')
            namenode = doc.createElement('ci')
            name = doc.createTextNode(function.getName())
            namenode.appendChild(name)
            topnode.appendChild(namenode)
            bottom_node = doc.createElement('apply')
            topnode.appendChild(bottom_node)
        else:
            topnode = doc.createElement('apply')
            bottom_node = topnode
        return bottom_node, topnode

    def setFunction(self, function):
        doc = self.ownerDocument
        topnode, funcnode = self.putName(function)
        for tag in self.functionMap.keys():
            if self.functionMap[tag][1].providedBy(function):
                func = doc.createElement(tag)
                break
        else:
            raise TypeError("%s is of an unknown function type" % function)
        for key in function.getAttributes().keys():
            func.setAttribute(key, function.getAttributes()[key])
        topnode.appendChild(func)
        for arg in function.getArguments():
            if isinstance(arg, (unicode, str)):
                argnode = doc.createElement('ci')
            elif isinstance(arg, (int, float)):
                argnode = doc.createElement('cn')
            else:
                raise TypeError("I don't know how to deal with the type of %s" % arg)
            argnode.appendChild(doc.createTextNode(str(arg)))
            topnode.appendChild(argnode)
        return funcnode

    def readFunction(self, **kwargs):
        topnode = self.getFirstChildByTagName('declare', 'apply')
        if topnode.tagName == 'declare':
            name = getText(topnode.getFirstChildByTagName('ci'))
            topnode = topnode.getFirstChildByTagName('apply')
        else:
            name =  None
        func = topnode.getFirstChildByTagName(*self.functionMap.keys())
        attrs = {}
        for node in [func.attributes.item(i) for i in \
                range(func.attributes.length)]:
            attrs[node.name] = node.value
        args = []
        for node in topnode.getChildByTagName('ci', 'cn'):
            if node.tagName == "ci":
               args.append(getText(node))
            elif node.tagName == "cn":
                arg = getText(node).strip()
                if arg.isdigit():
                    args.append(int(arg))
                else:
                    args.append(float(arg))
        return self.functionMap[func.tagName][0](args,
                attributes=attrs, name=name, **kwargs)


class InputElement(BaseElement, DescElementMixin):
    """A class for the input-data node in a TTML doc."""

    order = ['author', 'desc', 'aref', 'variables', 'domains',
            'constraints']


class VariablesElement(BaseElement):
    """A class for the variables node in a TTML doc."""

    order = [ 'declare', 'attrset' ]

    def objectifyVariableSet(self, set):
        """Parser with super cow powers for TTML sets."""
        members = ()
        element_attr_names = ()
        element_attrs = ()
        attr_names = ()
        attrs = ()
        # get the name and find out if we are dealing with a
        # simple or attributed set
        if set.tagName == 'declare':
            base_set_name = getText(set.getFirstChildByTagName('ci'))
            setnode = set.getFirstChildByTagName('set')
        if set.tagName == 'attrset':
            sets = set.getChildByTagName('declare')
            mainset = sets.next()
            base_set_name = getText(mainset.getFirstChildByTagName('ci'))
            setnode = mainset.getFirstChildByTagName('vector')
        # Parse the first set, looking for element arrtibutes
        element_attr_dict = {}
        for i in setnode.getChildByTagName('ci'):
            member = getText(i)
            members = members + (member, )
            element_attr_dict[member] = {}
            for node in [i.attributes.item(j) for j in range(i.attributes.length)]:
                    if node.name not in element_attr_names:
                        element_attr_names = element_attr_names + (node.name, )
                    element_attr_dict[member][node.name] = node.value
        # compile the element tuple from the temporary dictionary
        for i in element_attr_names:
            ele_attrs = ()
            for j in members:
                try:
                    ele_attr = element_attr_dict[j][i]
                except KeyError:
                    ele_attr = None
                ele_attrs = ele_attrs + (ele_attr, )
            element_attrs = element_attrs + (ele_attrs, )
        # If it is an attributed set, parse the attributes
        if set.tagName == 'attrset':
            for i in sets:
                attr_names = attr_names +\
                        (getText(i.getFirstChildByTagName('ci')), )
                vec = i.getFirstChildByTagName('vector')
                tmpattr = ()
                for j in vec.getChildByTagName('ci'):
                    tmpattr = tmpattr + (getText(j), )
                attrs = attrs + (tmpattr, )
        return data.VariableSet(base_set_name, members, attr_names, attrs,
                element_attr_names, element_attrs)

    def nodifyVariableSet(self, attrset):
        """Convert an object implementing IAttributedSet to a DOM node."""
        doc = self.ownerDocument
        # Find out what kind of set we are dealing with
        if not attrset.getAttributeNames():
            # Simple Set
            setnode = doc.createElement('declare')
            name = doc.createElement('ci')
            name.appendChild(doc.createTextNode(attrset.getBaseSetName()))
            setnode.appendChild(name)
            set = doc.createElement('set')
            for j in attrset.getBaseSetMembers():
                name = doc.createElement('ci')
                subname = doc.createTextNode(j)
                name.appendChild(subname)
                for i in attrset.getElementAttributeNames():
                    if attrset.getElementAttribute(i, j) != None:
                        name.setAttribute(i, attrset.getElementAttribute(i, j))
                set.appendChild(name)
            setnode.appendChild(set)
        else:
            # Attributed Set
            setnode = doc.createElement('attrset')
            # main set
            var = doc.createElement('declare')
            name = doc.createElement('ci')
            name.appendChild(doc.createTextNode(attrset.getBaseSetName()))
            var.appendChild(name)
            vector = doc.createElement('vector')
            for j in attrset.getBaseSetMembers():
                name = doc.createElement('ci')
                subname = doc.createTextNode(j)
                name.appendChild(subname)
                for i in attrset.getElementAttributeNames():
                    if attrset.getElementAttribute(i, j) != None:
                        name.setAttribute(i, attrset.getElementAttribute(i, j))
                vector.appendChild(name)
            var.appendChild(vector)
            setnode.appendChild(var)
            # attributes
            for i in attrset.getAttributeNames():
                var = doc.createElement('declare')
                name = doc.createElement('ci')
                name.appendChild(doc.createTextNode(i))
                var.appendChild(name)
                vector = doc.createElement('vector')
                for j in attrset.getBaseSetMembers():
                    name = doc.createElement('ci')
                    attribute = attrset.getAttribute(i, j)
                    name.appendChild(doc.createTextNode(attribute))
                    vector.appendChild(name)
                var.appendChild(vector)
                setnode.appendChild(var)
        return setnode

    def popVariableSet(self):
        setnode = self.getFirstChildByTagName('declare', 'attrset')
        varobj = self.objectifyVariableSet(setnode)
        self.removeChild(setnode)
        return varobj

    def setVariableSet(self, variables):
        doc = self.ownerDocument
        setnode = self.nodifyVariableSet(variables)
        self.removeAll('attrset')
        self.removeAll('declare')
        self.orderedInsert(setnode)


class DomainsElement(BaseElement):
    """A class for the domains node in a TTML doc."""

    order = [ 'time', 'duration', 'attrset', 'declare', 'domainsvar' ]

    def objectifyDomain(self, set):
        members = ()
        attr_names = ()
        attrs = ()
        # get the name and find out if we are dealing with a
        # simple or attributed set
        if set.tagName == 'declare':
            base_set_name = getText(set.getFirstChildByTagName('ci'))
            setnode = set.getFirstChildByTagName('set')
        if set.tagName == 'attrset':
            sets = set.getChildByTagName('declare')
            mainset = sets.next()
            base_set_name = getText(mainset.getFirstChildByTagName('ci'))
            setnode = mainset.getFirstChildByTagName('vector')
        # Parse the first set, looking for element arrtibutes
        element_attr_dict = {}
        for i in setnode.getChildByTagName('ci'):
            member = getText(i)
            members = members + (member, )
        # If it is an attributed set, parse the attributes
        if set.tagName == 'attrset':
            for i in sets:
                attr_names = attr_names +\
                        (getText(i.getFirstChildByTagName('ci')), )
                vec = i.getFirstChildByTagName('vector')
                tmpattr = ()
                for j in vec.getChildByTagName('ci'):
                    tmpattr = tmpattr + (getText(j), )
                attrs = attrs + (tmpattr, )
        # return everything we found
        return data.Domain(base_set_name, members, attr_names, attrs)

    def nodifyDomain(self, attrset):
        """Convert an object implementing IAttributedSet to a DOM node."""
        # Find out what kind of set we are dealing with
        doc = self.ownerDocument
        if not attrset.getAttributeNames():
            # Simple Set
            setnode = doc.createElement('declare')
            name = doc.createElement('ci')
            name.appendChild(doc.createTextNode(attrset.getBaseSetName()))
            setnode.appendChild(name)
            set = doc.createElement('set')
            for j in attrset.getBaseSetMembers():
                name = doc.createElement('ci')
                subname = doc.createTextNode(j)
                name.appendChild(subname)
                set.appendChild(name)
            setnode.appendChild(set)
        else:
            # Attributed Set
            setnode = doc.createElement('attrset')
            # main set
            var = doc.createElement('declare')
            name = doc.createElement('ci')
            name.appendChild(doc.createTextNode(attrset.getBaseSetName()))
            var.appendChild(name)
            vector = doc.createElement('vector')
            for j in attrset.getBaseSetMembers():
                name = doc.createElement('ci')
                name.appendChild(doc.createTextNode(j))
                vector.appendChild(name)
            var.appendChild(vector)
            setnode.appendChild(var)
            # attributes
            for i in attrset.getAttributeNames():
                var = doc.createElement('declare')
                name = doc.createElement('ci')
                name.appendChild(doc.createTextNode(i))
                var.appendChild(name)
                vector = doc.createElement('vector')
                for j in attrset.getBaseSetMembers():
                    name = doc.createElement('ci')
                    attribute = attrset.getAttribute(i, j)
                    name.appendChild(doc.createTextNode(attribute))
                    vector.appendChild(name)
                var.appendChild(vector)
                setnode.appendChild(var)
        return setnode

    def popDomain(self):
        node = self.getFirstChildByTagName('declare', 'attrset')
        domain = self.objectifyDomain(node)
        self.removeChild(node)
        return domain

    def appendDomain(self, domain):
        setnode = self.nodifyDomain(domain)
        self.orderedInsert(setnode, onlyone=False)

    def popDomainsvar(self):
        domainsvar = self.getFirstChildByTagName('domainsvar')
        declare = domainsvar.getFirstChildByTagName('declare')
        name = declare.getFirstChildByTagName('ci')
        namestr = getText(name)
        apply = declare.getFirstChildByTagName('apply')
        func = apply.getFirstChildByTagName('cartesianproduct')
        domains = [getText(node) for node in apply.getChildByTagName('ci')]
        domainsvarobj = data.Domainsvar(namestr, domains)
        self.removeChild(domainsvar)
        return domainsvarobj

    def setDomainsvar(self, domainsvar):
        doc = self.ownerDocument
        if interfaces.IDomainsvar.providedBy(domainsvar):
            domainsvarnode = doc.createElement('domainsvar')
            setnode = doc.createElement('declare')
            name = doc.createElement('ci')
            name.appendChild(doc.createTextNode(
                domainsvar.getDomainsvarName()))
            setnode.appendChild(name)
            set = doc.createElement('apply')
            func = doc.createElement('cartesianproduct')
            set.appendChild(func)
            for j in domainsvar.getDomains():
                name = doc.createElement('ci')
                name.appendChild(doc.createTextNode(j))
                set.appendChild(name)
            setnode.appendChild(set)
            domainsvarnode.appendChild(setnode)
        else:
            raise TypeError("Object must implement"
                    " IDomainsvar", domainsvar)
        self.orderedInsert(domainsvarnode)


class TimeDomainElement(BaseElement, FunctionMixin):

    order = [ 'declare', 'apply' ]
    functionMap = {
            'tmatrix': (data.TmatrixTimeDomain,
                    interfaces.ITmatrixTimeDomain)}

    def convertDurationStrToObj(self, str):
        """Convert a text representation of a duration into a timedelta object.
        """
        # XXX - This is horrible and can be done a lot better
        day = hr = min = sec = 0
        num = ''
        for char in str:
            if char == 'P' or char == 'T':
                state = char
                continue
            if char in '1234567890':
                num = num + char
                continue
            if state == 'P':
                if char == 'Y' or char == 'M':
                    raise ValueError("Time intervals with years or months are"
                            "unsupported")
                elif char == 'D':
                    day = int(num)
            if state == 'T':
                if char == 'H':
                    hr = int(num)
                elif char == 'M':
                    min = int(num)
                elif char == 'S':
                    sec = int(num)
            num = ''
        return datetime.timedelta(days=day, hours=hr, minutes=min, seconds=sec)

    def convertTimeStrToObj(self, timestr):
        """Convert a text representation of a date/time to datetime."""
        # XXX - this is a _really_ horrible conversion and should be done
        # better, it will bork on most things
        year = datetime.MINYEAR
        month = 1
        day = 1
        hr = min = sec = 0
        num = ''
        state = 1
        for char in timestr + 'f':
            if char in '1234567890':
                num = num + char
                continue
            if state == 1:
                year = int(num)
            elif state == 2:
                month = int(num)
            elif state == 3:
                day = int(num)
            elif state == 4:
                hr = int(num)
            elif state == 5:
                min = int(num)
            elif state == 6:
                sec = int(num)
            state = state + 1
            num = ''
        return datetime.datetime(year, month, day, hour=hr, minute=min,
                second=sec)

    def popTimeDomain(self):
        """This function returns the existing time domain.

        This function decides which time domain specific function to call to
        read the time domain.
        """
        try:
            timefunc = self.getFirstChildByTagName('declare')
            timeset = timefunc.getFirstChildByTagName('set', 'apply')
        except IndexError:
            timeset = self.getFirstChildByTagName('set', 'apply')
            timefunc = timeset
        if timeset.tagName == 'set':
            ans = self.readTimeDomainBasic()
        elif timeset.tagName == 'apply':
            ans = self.readFunction()
        else:
            raise TypeError("Whoops! Can't read the time domain of an "
                    "unknown time domain")
        self.removeChild(timefunc)
        return ans

    def setTimeDomain(self, timeDomain):
        self.removeAll(self.order)
        if interfaces.ITimeDomain.providedBy(timeDomain):
            self.setTimeDomainBasic(timeDomain)
        else:
            self.orderedInsert(self.setFunction(timeDomain))

    def setTimeDomainBasic(self, timeobj):
        """Set a time domain from an object implementing ITimeDomain."""
        doc = self.ownerDocument
        domain = doc.createElement('declare')
        namenode = doc.createElement('ci')
        namenode.appendChild(doc.createTextNode(timeobj.getName()))
        domain.appendChild(namenode)
        domainfunc = doc.createElement('set')
        if timeobj.isDiscrete():
            domainfunc.setAttribute('interval', 'discrete')
        for i in timeobj.getIntervals():
            interval = doc.createElement('cn')
            interval.appendChild(doc.createTextNode(i[0].isoformat()))
            sep  = doc.createElement('sep')
            interval.appendChild(sep)
            timedelta = 'P' + str(i[1].days) + 'DT'\
                    + str(i[1].seconds) + 'S'
            interval.appendChild(doc.createTextNode(timedelta))
            domainfunc.appendChild(interval)
        domain.appendChild(domainfunc)
        self.orderedInsert(domain)

    def readTimeDomainBasic(self):
        """Returns an object representing the time domain.

        The object will implement ITimeDomain."""
        declaration = self.getFirstChildByTagName('declare')
        name = getText(declaration.getFirstChildByTagName('ci'))
        timeset = declaration.getFirstChildByTagName('set')
        if timeset.getAttribute('interval') == 'discrete':
            isdiscrete = True
        else:
            isdiscrete = False
        intervals = []
        for node in timeset.getChildByTagName('cn'):
            timestr = node.childNodes[0].data
            durationstr = node.childNodes[2].data
            intervals.append((self.convertTimeStrToObj(timestr),
                self.convertDurationStrToObj(durationstr)))
        domain = data.TimeDomain(name, intervals, isdiscrete=isdiscrete)
        return domain


class ConstraintsElement(BaseElement):
    """A class for the constraints node in a TTML doc."""

    order = [ 'classifiers', 'hard', 'soft' ]


class CoiElement(BaseElement, FunctionMixin):
    """Class for a coi node"""

    functionMap = {
            'notsame': (data.NotSameConstraint,
                    interfaces.INotSameConstraint),
            'nooverlap': (data.NoOverlapConstraint,
                    interfaces.INoOverlapConstraint),
            'preset': (data.PresetConstraint,
                    interfaces.IPresetConstraint),
            'exclude': (data.ExcludeConstraint,
                    interfaces.IExcludeConstraint),
            'ordering': (data.OrderingConstraint,
                    interfaces.IOrderingConstraint),
            'eventspr': (data.EventSprConstraint,
                    interfaces.IEventSprConstraint),
            'fullspr': (data.FullSprConstraint,
                    interfaces.IFullSprConstraint),
            'freespr': (data.FreeSprConstraint,
                    interfaces.IFreeSprConstraint),
            'attrcomp': (data.AttrCompConstraint,
                    interfaces.IAttrCompConstraint),
            'resnoclash': (data.ResNoClashConstraint,
                    interfaces.IResNoClashConstraint),
            'chksum': (data.ChksumConstraint,
                    interfaces.IChksumConstraint)}


class HardOrSoftConstraintsElement(BaseElement):
    """A class for a hard or soft constraints sub node in a TTML doc."""

    order = [ 'coi' ]

    def appendConstraint(self, constraint):
        doc = self.ownerDocument
        coi = doc.createElement('coi')
        coi.appendChild(coi.setFunction(constraint))
        self.orderedInsert(coi, onlyone=False)

    def popConstraint(self):
        if self.tagName == 'hard':
            ishard = True
        else:
            ishard = False
        coi = self.getFirstChildByTagName('coi')
        func = coi.readFunction(ishard=ishard)
        self.removeChild(coi)
        return func


class ClassifiersElement(BaseElement):
    """A class for the classifiers node in a TTML doc."""

    order = [ 'rootcl' ]

    def objectifyClassifier(self, set):
        """A recursive writer od classifiers."""
        members = ()
        base_set_name = getText(set.getFirstChildByTagName('ci'))
        setnode = set.getFirstChildByTagName('set')
        element_attr_dict = {}
        for i in setnode.getChildByTagName('ci'):
            try:
                subset = i.getFirstChildByTagName('declare')
            except IndexError:
                subset = False
            if subset:
                member = self.objectifyClassifier(subset)
            else:
                member = getText(i)
            members = members + (member, )
        return data.Classifier(base_set_name, members)

    def nodifyClassifier(self, classifier):
        """Revursive classifier nodification"""
        doc = self.ownerDocument
        setnode = doc.createElement('declare')
        name = doc.createElement('ci')
        name.appendChild(doc.createTextNode(classifier.getBaseSetName()))
        setnode.appendChild(name)
        set = doc.createElement('set')
        for j in classifier.getBaseSetMembers():
            name = doc.createElement('ci')
            if interfaces.IClassifier.providedBy(j):
                subname = self.nodifyClassifier(j)
            else:
                subname = doc.createTextNode(j)
            name.appendChild(subname)
            set.appendChild(name)
        setnode.appendChild(set)
        return setnode

    def popRootClassifier(self):
        node = self.getFirstChildByTagName('rootcl')
        decl = node.getFirstChildByTagName('declare')
        classifier = self.objectifyClassifier(decl)
        self.removeChild(node)
        return classifier

    def appendRootClassifier(self, classifier):
        if not interfaces.IClassifier.providedBy(classifier):
            raise TypeError("%s must implemet %s"
                    % (classifier, interfaces.IClassifier))
        doc = self.ownerDocument
        rootcl = doc.createElement('rootcl')
        rootcl.appendChild(self.nodifyClassifier(classifier))
        self.orderedInsert(rootcl, onlyone=False)


class OutputElement(BaseElement):

    order = [ 'li' ]


class TestResultsElement(BaseElement):

    order = [ 'result' ]


class ResultElement(BaseElement, DescElementMixin):

    order = [ 'author', 'pref', 'desc', 'best', 'penalty', 'eval' ]


class TimeTablingElement(BaseElement):
    """A class for the classifiers node in a TTML doc."""

    order = [ 'input-data', 'output', 'test-results' ]

#
# Document Objects
#

class DOMImplementation(minidom.DOMImplementation):
    """A TTML DOMImplementation."""

    def _create_document(self):
        return Document()


class Document(minidom.Document):
    """A TTML Document Object.

    Here we hook into the DOM model by sub classing the minidom Document
    class and replacing the basic Element class with our subclasses.
    """

    implementation = DOMImplementation()

    def _chooseElementType(self, tagName):
        if tagName == 'input-data':
            return InputElement
        elif tagName == 'variables':
            return VariablesElement
        elif tagName == 'domains':
            return DomainsElement
        elif tagName == 'time':
            return TimeDomainElement
        elif tagName == 'constraints':
            return ConstraintsElement
        elif tagName == 'hard' or tagName == 'soft':
            return HardOrSoftConstraintsElement
        elif tagName == 'coi':
            return CoiElement
        elif tagName == 'classifiers':
            return ClassifiersElement
        elif tagName == 'output':
            return OutputElement
        elif tagName == 'test-results':
            return TestResultsElement
        elif tagName == 'result':
            return ResultElement
        elif tagName == 'time-tabling':
            return TimeTablingElement
        else:
            return BaseElement

    def createElementNS(self, namespaceURI, qualifiedName):
        """Re-definition of createElementNS using TTML element subclasses"""
        Element = self._chooseElementType(qualifiedName)
        prefix, localName = minidom._nssplit(qualifiedName)
        e = Element(qualifiedName, namespaceURI, prefix)
        e.ownerDocument = self
        return e

    def createElement(self, tagName):
        """Re-definition of createElement using TTML element subclasses"""
        Element = self._chooseElementType(tagName)
        e = Element(tagName)
        e.ownerDocument = self
        return e


class DocumentWrapper:
    """An object implementing IDoc.

    This object implements IDoc by providing a mapping between IDoc methods
    and the  methods attached to the DOM tree.

    Note that care is taken to not create nodes on reading (i.e. 'get'
    methods). It is a feature that these will raise an error if the node does
    not exist.
    """

    # Hooks for unit tests
    document_hook = Document

    implements(interfaces.IDoc)
    mapping = {'input-data': ['input-data'],
            'variables' : [ 'input-data', 'variables' ],
            'domains': ['input-data', 'domains'],
            'hard': ['input-data', 'constraints', 'hard' ],
            'soft': ['input-data', 'constraints', 'soft' ],
            'classifiers': ['input-data', 'constraints', 'classifiers'],
            'time': ['input-data', 'domains', 'time']}

    def __init__(self, doc):
        if not isinstance(doc, self.document_hook):
            raise TypeError("Document %s must be a subclass of %s"
                    % (doc, self.document_hook))
        self.doc = doc
        self.doc_node = doc.documentElement

    def getMappedNode(self, mapindex, create=False):
        node = self.doc_node
        for i in self.mapping[mapindex]:
            try:
                tmp = node.getFirstChildByTagName(i)
            except IndexError:
                if create:
                    tmp = self.doc.createElement(i)
                    node.orderedInsert(tmp)
                else:
                    raise IndexError("Document does not contain node path %s"
                            % self.mapping[mapindex])
            node = tmp
        return node

    def addObject(self, obj):
        """Routes objects to the correct nodes in the dom tree."""
        if interfaces.IInputDescription.providedBy(obj):
            node = self.getMappedNode('input-data', create=True)
            return node.setDescriptive(obj)
        elif interfaces.IVariableSet.providedBy(obj):
            node = self.getMappedNode('variables', create=True)
            return node.setVariableSet(obj)
        elif interfaces.ITimeDomain.providedBy(obj):
            node = self.getMappedNode('time', create=True)
            return node.setTimeDomain(obj)
        elif interfaces.ITmatrixTimeDomain.providedBy(obj):
            node = self.getMappedNode('time', create=True)
            return node.setTimeDomain(obj)
        elif interfaces.IDomain.providedBy(obj):
            node = self.getMappedNode('domains', create=True)
            return node.appendDomain(obj)
        elif interfaces.IDomainsvar.providedBy(obj):
            node = self.getMappedNode('domains', create=True)
            return node.setDomainsvar(obj)
        elif interfaces.IClassifier.providedBy(obj):
            node = self.getMappedNode('classifiers', create=True)
            return node.appendRootClassifier(obj)
        elif interfaces.IConstraintFunction.providedBy(obj):
            if obj.isHard():
                node = self.getMappedNode('hard', create=True)
            else:
                node = self.getMappedNode('soft', create=True)
            return node.appendConstraint(obj)
        else:
            raise TypeError("%s does not implement a data object interface"
                    % obj)

    def popInputDescriptive(self):
        node = self.getMappedNode('input-data')
        return node.popDescriptive()

    def popVariableSet(self):
        node = self.getMappedNode('variables')
        return node.popVariableSet()

    def popTimeDomain(self):
        node = self.getMappedNode('time')
        return node.popTimeDomain()

    def popDomain(self):
        node = self.getMappedNode('domains')
        return node.popDomain()

    def popDomainsvar(self):
        node = self.getMappedNode('domains')
        return node.popDomainsvar()

    def popRootClassifier(self):
        node = self.getMappedNode('classifiers')
        return node.popRootClassifier()

    def popHardConstraint(self):
        node = self.getMappedNode('hard')
        return node.popConstraint()

    def popSoftConstraint(self):
        node = self.getMappedNode('soft')
        return node.popConstraint()


class DocumentFactory:

    implementation_hook = DOMImplementation()
    expatbuilder.theDOMImplementation = implementation_hook
    builder_hook = expatbuilder.ExpatBuilder
    documentwrapper_hook = DocumentWrapper
    fileopen_hook = open

    implements(interfaces.IDocumentFactory)

    def parse(self, file):
    #    events.expandNode(rootNode)
        builder = self.builder_hook()
    #    if isinstance(file, str):
    #    builder = expatbuilder.ExpatBuilderNS()
    #    from xml.dom.pulldom import DOMEventStream
    #    builder = DOMEventStream()
    #    builder.pulldom.documentFactory = DOMImplementation()
        if isinstance(file, str):
            fp = self.fileopen_hook(file, 'rb')
            try:
                result = builder.parseFile(fp)
            finally:
                fp.close()
        else:
            result = builder.parseFile(file)
        return self.documentwrapper_hook(result)

    def create(self, schema='http://minittml.alioth.debian.org/schema'):
        doctype = self.implementation_hook.createDocumentType(
                'time-tabling', '', '')
        doc = self.implementation_hook.createDocument(schema, 'time-tabling',
                doctype)
        return self.documentwrapper_hook(doc)
